import 'package:contentsquare/src/modules/screengraph/render_object_extension.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';

/// An enum describing a boolean with a third value: unknown
enum MaybeResult {
  /// true
  yes,

  /// false
  no,

  /// unknown
  unknown,
}

/// An extension on [Element] that provides some utility methods
extension ElementExtension on Element {
  /// Returns the [RenderObject] of the given [Element] if any
  ///
  /// The difference between this and [Element.renderObject] is that
  /// this does not walk the tree to get the 1st [RenderObject] but only
  /// returns the [RenderObject] attached directly to the current [Element]
  /// which only happens if it's a [RenderObjectElement]
  RenderObject? get directRenderObject {
    final _this = this;
    if (_this is RenderObjectElement) {
      return _this.renderObject; // Not null !
    }

    return null;
  }

  /// Whether this element contains the [globalPosition]
  ///
  /// Returns [MaybeResult.unknown] if we can't say if it contains the
  /// [globalPosition] (because this [Element] is not associated with a
  /// [RenderObject])
  MaybeResult containsGlobalPosition(Offset globalPosition) {
    final _directRenderObject = directRenderObject;

    // If this element has no direct render object, we can't say if it contains
    // the globalPosition
    if (_directRenderObject == null) {
      return MaybeResult.unknown;
    }

    final globalFrame = _directRenderObject.globalFrame;

    if (globalFrame != null && globalFrame.contains(globalPosition)) {
      return MaybeResult.yes;
    }

    return MaybeResult.no;
  }

  /// Returns the name of the [Widget] associated with this element
  String get widgetName => widget.runtimeType.toString();

  /// Calls the [visit] function on each child of this element until the [visit]
  /// function returns true.
  ///
  /// [elementTVP] must be the TVP to this element and is used to create the TVP
  /// of the children.
  /// See [_createTVPFromParent] to understand how childPath is built.
  void visitChildrenFromTVP({
    required String elementTVP,
    required bool Function(Element childElement, String childPath) visit,
  }) {
    // [RenderSliverMultiBoxAdaptor] have a special treatment because they
    // have several children which can be recycled.
    //
    // Instead of using an index based on the type, we use the index value of
    // each child directly (using the fact that they are [RenderIndexedSemantics],
    // which holds an index). This ensures that the the TVP is consistent across
    // different scroll positions.
    final isRenderSliverMultiBoxAdaptor =
        renderObject is RenderSliverMultiBoxAdaptor;

    // The [typeCounts] map contains the number of times each [RenderObject] type
    // appears in the childrenElement, which is used to compute the path of each
    // [RenderObjectNode]
    final Map<Type, int> typeCounts = <Type, int>{};
    var _shouldVisit = true;
    visitChildren((childElement) {
      if (_shouldVisit) {
        final widgetType = childElement.widget.runtimeType;

        // Update the count based on the widget type
        final count = (typeCounts[widgetType] ?? 0) + 1;
        typeCounts[widgetType] = count;

        final int index;
        if (isRenderSliverMultiBoxAdaptor) {
          // Use the index of the [RenderIndexedSemantics] so that the TVP is
          // consistent across different scroll positions
          final childRenderObject = childElement.renderObject;
          if (childRenderObject is RenderIndexedSemantics) {
            index = childRenderObject.index;
          } else {
            index = count - 1;
          }
        } else {
          index = count - 1;
        }

        final childPath = _createTVPFromParent(
          parentTVP: elementTVP,
          childElement: childElement,
          childIndex: index,
        );

        _shouldVisit = !visit(childElement, childPath);
      }
    });
  }
}

/// The TVP of [childElement] is the concatenation of:
///   1. The TVP of the parent
///   2. The runtimeType of the widget of the [childElement]
///   3. The index of [childElement] in the parent's children list of the same type
String _createTVPFromParent({
  required String parentTVP,
  required Element childElement,
  required int childIndex,
}) {
  return '$parentTVP>${childElement.widgetName}:eq($childIndex)';
}
